/*
 * CO2mini access program
 * 2017-01-07 Kouichi Hirabayashi
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <hidapi/hidapi.h>
#include <linux/uhid.h>
#include <linux/hiddev.h>

#define MANUFACTURER_ID 0x4d9		// Holtek
#define PRODUCT_ID	0xa052		// USB-zyTemp

typedef unsigned char co2mon_data_t[8];

void swap_char(unsigned char *, unsigned char *);
void decode_buf(co2mon_data_t, co2mon_data_t, co2mon_data_t);
void error(char *fmt, ...);

co2mon_data_t magic_table = { 0 };
co2mon_data_t data = { 0 };
co2mon_data_t val;

char *cmd;

int main(int argc, char **argv)
{
  char *s;
  hid_device *dev;
  int n, cksum, w;
  double Tamb = -1e3;
  int CntR = -100;
  cmd = argv[0];

  if (hid_init() < 0)
	error("hid_init() error");

  if ((dev = hid_open(MANUFACTURER_ID, PRODUCT_ID, NULL)) == NULL)
	error("hid_open(0x%x, 0x%x) error", MANUFACTURER_ID, PRODUCT_ID);

  if (hid_send_feature_report(dev, magic_table, sizeof(co2mon_data_t)) < 0)
	perror("hid_send_feature_report()");

  do {
	if ((n = hid_read_timeout(dev, data, sizeof(data), 5000)) < 0)
		perror("hid_read_timeout()");
	if (n != sizeof(data))
		fprintf(stderr, "read %d bytes, expected %d bytes.\n",
			n, (int)sizeof(data));
	decode_buf(val, data, magic_table);
	if (val[4] != 0x0D)
		fprintf(stderr, "Unexpected data (%02X) != 0x0D.\n", val[4]);
	cksum = (val[0] + val[1] + val[2]) & 0xFF;
	if (cksum != val[3])
		fprintf(stderr, "checksum error (%02X) != %02X\n",
			cksum, val[3]);
	w = (val[1] <<8) + val[2];
	switch (val[0]) {
	case 'B':	// 0x42
		Tamb = (double)w / 16.0 - 273.15;
		break;
	case 'P':	// 0x50
		CntR = (int)w;
		break;
	default:
		// ignore
		break;
	}
  } while (Tamb == -1e3 || CntR == -100);
  printf("%.1f C %d ppm\n", Tamb, CntR);

  hid_close(dev);
  hid_exit();

  exit(0);
}

void swap_char(unsigned char *a, unsigned char *b)
{
  unsigned char tmp = *a;

  *a = *b;
  *b = tmp;
}

void decode_buf(co2mon_data_t val, co2mon_data_t buf, co2mon_data_t magic_table)
{
  const unsigned char magic_word[8] = "Htemp99e";
  int i;
  unsigned char tmp;

  swap_char(&buf[0], &buf[2]);
  swap_char(&buf[1], &buf[4]);
  swap_char(&buf[3], &buf[7]);
  swap_char(&buf[5], &buf[6]);

  tmp = (buf[7] << 5);
  val[7] = (buf[6] << 5) | (buf[7] >> 3);
  val[6] = (buf[5] << 5) | (buf[6] >> 3);
  val[5] = (buf[4] << 5) | (buf[5] >> 3);
  val[4] = (buf[3] << 5) | (buf[4] >> 3);
  val[3] = (buf[2] << 5) | (buf[3] >> 3);
  val[2] = (buf[1] << 5) | (buf[2] >> 3);
  val[1] = (buf[0] << 5) | (buf[1] >> 3);
  val[0] = tmp | (buf[0] >> 3);
  for (i = 0; i < 8; ++i) {
	val[i] -= (magic_word[i] << 4) | (magic_word[i] >> 4);
  }
}

dump(s, n) char *s; int n;
{
  int i;

  for (i = 0; i < n; i++) {
	printf(" %02x", s[i] & 0xff);
  }
  printf("\n");
  return 0;
}

void error(char *fmt, ...)
{
  va_list argp;

  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
  fprintf(stderr, "\n");
  exit(1);
}
